/**
 * @file
 * Defines behaviors to support Opayo Pi 3D Secure authentication
 *
 */
(function (Drupal, once) {

  'use strict';

  /**
   * Attaches the Opayo3DSecure behavior.
   *
   * @type {Drupal~behavior}
   *
   * @prop {Drupal~behaviorAttach} attach
   *   Attaches the OpayoCheckout behavior.
   */
  Drupal.behaviors.Opayo3DSecure = {
    attach: function (context, settings) {
      once('Opayo3DSecure', '#commerce-opayo-3d-iframe', context).forEach((ele) => {

        // Cannot proceed at this stage until 3D secure authentication has completed
        let proceedButton = document.querySelector('[data-drupal-selector="edit-actions-next"]');
        proceedButton.disabled = true;

        // Define an Ajax command that can be called via a PHP response 'Command' so that we can process the response to the 3D secure progress request and act on any included 'instruction'
        Drupal.AjaxCommands.prototype.opayo3DSecureResult = function (ajax, response, status) {
          if (response.settings.hasOwnProperty("result")) {
            if (response.settings.result != "inprogress")
            {
              // There is a resolution
              let message = '';
              if (response.settings.hasOwnProperty("message")) {
                message = response.settings.message;
              }
              // Cancel the timer
              let timerRef = null;
              if (settings.commerceopayopi.opayo_3dsecure.timerRef) {
                timerRef = settings.commerceopayopi.opayo_3dsecure.timerRef;
              }
              // Process the result
              process3DSecureResult(result, message, timerRef);
            }
            // Do nothing if the authentication is still 'in progress'
          }
        };

        // Timer to check for 3D Authentication results
        let orderId = document.querySelector('#commerce-opayo-order-id').value;
        let transactionId = document.querySelector('#commerce-opayo-transaction-id').value;
        if (!settings.hasOwnProperty('commerceopayopi'))
          settings.commerceopayopi = new Object();
        if (!settings.commerceopayopi.hasOwnProperty('opayo_3dsecure'))
          settings.commerceopayopi.opayo_3dsecure = new Object();
        settings.commerceopayopi.opayo_3dsecure.timerRef = start3DResultTimer(10000, transactionId, orderId);

        // Add event listener for messages from the 3D Secure iFrame
        window.addEventListener('message', function (e) {
          const message = JSON.parse(e.data);
          // Result feedback for the back-end checkout pane
          document.querySelector('#commerce-opayo-3d-result').value = message.message;

          // Process the result
          let messageForUser = '';
          if (message.hasOwnProperty('messageforuser') && message.messageforuser != '') {
            messageForUser = message.messageforuser;
          }
          process3DSecureResult(message.result, messageForUser, settings.commerceopayopi.opayo_3dsecure.timerRef);
        });

      }, context);
    }
  }

  function process3DSecureResult(result, message, timerRef) {
    if (timerRef != null) {
      // Cancel the timer that keeps checking for results
      clearInterval(timerRef);
    }
    // 3D Secure authentication completed, submit the form associated with the 'threedsecure' checkout flow state to continue to the next appropriate step
    document.querySelector('#commerce-opayo-3d-result').value = result;
    if (result == 'success') {
      // Proceed to the next page of checkout (should be the 'complete' stage)
      let proceedButton = document.querySelector('[data-drupal-selector="edit-actions-next"]');
      proceedButton.disabled = false;
      let checkoutForm = document.querySelector('form[id^="commerce-checkout"]');
      checkoutForm.submit();
    }
    else {
      // Check if there is a message to display to the user
      if (message != '') {
        const drupalMessage = new Drupal.Message();
        drupalMessage.add(message, {type: 'warning'});
        // Submit feedback message with the form so that it can be displayed if there's a redirect
        document.querySelector('#commerce-opayo-3d-user-message').value = message;
      }
      let proceedButton = document.querySelector('[data-drupal-selector="edit-actions-next"]');
      proceedButton.disabled = false;
      let checkoutForm = document.querySelector('form[id^="commerce-checkout"]');
      checkoutForm.submit();
    }
  }

  function start3DResultTimer(msInterval, transactionId, orderId) {
    // Start an interval to check for 3D results - as a backup for the messaging with the iFrame
    let timerRef = setInterval((merchantSessionKey, orderId) => {
      const threeDSecureIframe = document.getElementById('commerce-opayo-3d-iframe');
      if (threeDSecureIframe) {
        // Still displaying the 3D secure iframe
        let ajaxObject = Drupal.ajax({
          url: '/commerce_opayo_pi/3dSecure_result/' + orderId + '/' + transactionId,
          base: false,
          element: false,
          progress: false
        });
        ajaxObject.execute();
      }
    }, msInterval, transactionId, orderId);
    return timerRef;
  }

})(Drupal, once);
