/**
 * @file
 * Defines behaviors to support Opayo Pi checkout using the 'drop-in form'
 *
 */
(function (Drupal, once) {

  'use strict';

  /**
   * Attaches the OpayoCheckout behavior.
   *
   * @type {Drupal~behavior}
   *
   * @prop {Drupal~behaviorAttach} attach
   *   Attaches the OpayoCheckout behavior.
   */
  Drupal.behaviors.OpayoPiOwnForm = {
    attach: function (context, settings) {
      once('OpayoPiOwnForm', '#card-identifier', context).forEach((ele) => {

        // Call the function that will deal with the merchant session key and associated expiries
        processMerchantSessionKey(context, settings);

        // Define an Ajax command that can be called via a PHP response 'Command' so that we can process a new merchant session key and act on any included 'instruction'
        Drupal.AjaxCommands.prototype.opayoMerchantSessionKey = function (ajax, response, status) {
          // Process the new merchant session key
          processMerchantSessionKey(context, settings, response.settings, true);

          if (response.settings.hasOwnProperty("instruction") && response.settings.instruction == "tokenize_and_submit") {
            // New merchant session key needs to be used to tokenize the card details and submit the checkout form
            callCardDetailTokenisation(context, settings);
          }
        };

        // Event handler for copying country change to telephone number
        // Actually, doesn't belong here - should move to module with IHB tweaks to the commerce behaviour
        let countryCodeElements = document.getElementsByName('payment_information[add_payment_method][billing_information][address][0][address][country_code]');
        if (countryCodeElements && countryCodeElements.length > 0) {
          countryCodeElements[0].addEventListener('change', function (e) {
            let countryCode = this.value;
            // TODO - change the telephone number's country - doable but fiddly I think
          }, false);
        }

        // Copy from first name and surname to cardholder unless the cardholder has been manually entered
        // Will need to keep track whether the 'cardholder name' has been modified by user input (in which case the event handler will leave it alone)
        context.querySelector('[data-card-details="cardholder-first-name"]').addEventListener('change', function (e) { copyCardholderName(e); }, false);
        context.querySelector('[data-card-details="cardholder-last-name"]').addEventListener('change', function (e) { copyCardholderName(e); }, false);
        context.querySelector('[data-card-details="cardholder-name"]').addEventListener('change', function (e) { copyCardholderName(e); }, false);
        let checkoutForm = context.querySelector('form[id^="commerce-checkout"]');
        let cardHolderNameHidden = document.createElement("input");
        cardHolderNameHidden.setAttribute("type", "hidden");
        cardHolderNameHidden.setAttribute("data-card-details", "cardholder-name-manual-entry");
        checkoutForm.appendChild(cardHolderNameHidden);

        // Add an event listener for the form submit button that will first contact Opayo to tokenise the card details, add the new 'card identifier' to the form and submit the form back to our server
        let checkoutFormButton = context.querySelector('#edit-actions-next');
        checkoutFormButton.addEventListener('click', function (e) {
          e.preventDefault(); // to prevent form submission

          // Fetch a merchant session key first by requesting the back-end to fetch and return one using AJAX
          // Result will
          let ajaxObject = Drupal.ajax({
            url: '/commerce_opayo_pi/merchantsessionkey/' + settings.commerceopayopi.opayoCheckout.orderId + '/tokenize_and_submit',
            base: false,
            element: false,
            progress: false
          });
          ajaxObject.execute();

        }, false);

        // Tidy up on submit
        context.querySelector('[type=submit]').addEventListener('submit', function (e) {
          try {
            if (settings.commerceopayopi.opayoCheckout.mskTimeout) {
              clearTimeout(settings.commerceopayopi.opayoCheckout.mskTimeout);
            }
          }
          catch { }
        }, false);
      }, context);
    }
  }

  function copyCardholderName(e) {
    const element = e.currentTarget;
    let dataCardDetails = element.getAttribute("data-card-details");
    if (dataCardDetails == "cardholder-first-name" || dataCardDetails == "cardholder-last-name") {
      if (!(document.querySelector('[data-card-details="cardholder-name-manual-entry"]').value))
      {
        // User has not typed into the cardholder name field, copy from first name and surname fields
        let firstName = document.querySelector('[data-card-details="cardholder-first-name"]').value;
        let lastName = document.querySelector('[data-card-details="cardholder-last-name"]').value;
        let cardholderName = firstName + (firstName.length > 0 && lastName.length > 0 ? ' ' : '') + lastName;
        document.querySelector('[data-card-details="cardholder-name"]').value = cardholderName;
      }
    }
    else if (dataCardDetails == "cardholder-name" && e.isTrusted) {
      // User modified cardholder name field
      document.querySelector('[data-card-details="cardholder-name-manual-entry"]').value = element.value;
    }
  }

  function getCardType (number) {
  const re = {
    electron: /^(4026|417500|4405|4508|4844|4913|4917)\d+$/,
    maestro: /^(5018|5020|5038|5612|5893|6304|6759|6761|6762|6763|0604|6390)\d+$/,
    dankort: /^(5019)\d+$/,
    interpayment: /^(636)\d+$/,
    unionpay: /^(62|88)\d+$/,
    visa: /^4[0-9]{12}(?:[0-9]{3})?$/,
    mastercard: /^(5[1-5][0-9]{14}|2(22[1-9][0-9]{12}|2[3-9][0-9]{13}|[3-6][0-9]{14}|7[0-1][0-9]{13}|720[0-9]{12}))$/,
    amex: /^3[47][0-9]{13}$/,
    diners: /^3(?:0[0-5]|[68][0-9])[0-9]{11}$/,
    discover: /^6(?:011|5[0-9]{2})[0-9]{12}$/,
    jcb: /^(?:2131|1800|35\d{3})\d{11}$/
  }
  for (var key in re) {
    if (re[key].test(number)) {
      return key
    }
  }
  return 'unknown'
}

  function processMerchantSessionKey(context, settings, newSettings = null, refresh = false) {
    if (newSettings != null) {
      // Update the merchant session key that we need to use to tokenise the card details
      settings.commerceopayopi.opayoCheckout.merchantSessionKey = newSettings.merchantSessionKey;
    }

    if (refresh) {
      if (settings.commerceopayopi.opayoCheckout.mskTimeout) {
        clearTimeout(settings.commerceopayopi.opayoCheckout.mskTimeout);
      }
    }

    // Make sure that the hidden form fields representing the merchant session key are up-to-date
    let drupalSettingsMSKSequenceNr = -1;
    let hiddenFormFieldsMSKSequenceNr = -1;
    let merchantSequenceKeyFormField = null;
    try {
      drupalSettingsMSKSequenceNr = settings.commerceopayopi.opayoCheckout.merchantSessionKeySequenceNr;
      merchantSequenceKeyFormField = context.querySelector('#merchant-session-key-sequence-nr');
      hiddenFormFieldsMSKSequenceNr = parseInt(merchantSequenceKeyFormField.value);
    }
    catch {}
    if (merchantSequenceKeyFormField != null && hiddenFormFieldsMSKSequenceNr < drupalSettingsMSKSequenceNr)
    {
      // Update the hidden form fields
      context.querySelector('#merchant-session-key').value = settings.commerceopayopi.opayoCheckout.merchantSessionKey;
      context.querySelector('#merchant-session-key-sequence-nr').value = settings.commerceopayopi.opayoCheckout.merchantSessionKeySequenceNr;
      context.querySelector('#merchant-session-key-expiry').value = settings.commerceopayopi.opayoCheckout.merchantSessionKeyExpiry;
    }
    else if (merchantSequenceKeyFormField != null && hiddenFormFieldsMSKSequenceNr > drupalSettingsMSKSequenceNr) {
      // Update the drupal settings
      settings.commerceopayopi.opayoCheckout.merchantSessionKey = context.querySelector('#merchant-session-key').value;
      settings.commerceopayopi.opayoCheckout.merchantSessionKeySequenceNr = context.querySelector('#merchant-session-key-sequence-nr').value;
      settings.commerceopayopi.opayoCheckout.merchantSessionKeyExpiry = context.querySelector('#merchant-session-key-expiry').value;
    }

    // Time remaining on merchant session key
    const skExpiry = Date.parse(newSettings == null ? settings.commerceopayopi.opayoCheckout.merchantSessionKeyExpiry : newSettings.merchantSessionKeyExpiry);
    const merchantSessionKey = newSettings == null ? settings.commerceopayopi.opayoCheckout.merchantSessionKey : newSettings.merchantSessionKey;
    const dateNow = new Date().getTime();
    if (skExpiry > dateNow) {
      let timeoutInMs = skExpiry - dateNow;
      // Start a timeout based on the merchant session key's validity
      let timeoutRef = startMerchantSessionKeyTimeout(timeoutInMs, merchantSessionKey, settings.commerceopayopi.opayoCheckout.orderId);
      settings.commerceopayopi.opayoCheckout.mskTimeout = timeoutRef;
    }
    else {
      // Request new merchant session key
      let ajaxObject = Drupal.ajax({
        url: '/commerce_opayo_pi/merchantsessionkey/' + settings.commerceopayopi.opayoCheckout.orderId,
        base: false,
        element: false,
        progress: false
      });
      ajaxObject.execute();

    }
  }


  function callCardDetailTokenisation(context, settings) {

    let cardNumber = context.querySelector('[data-card-details="card-number"]').value;
    settings.commerceopayopi.opayoCheckout.cardType = getCardType(cardNumber);
    settings.commerceopayopi.opayoCheckout.cardLastFourDigits = cardNumber.slice(-4);
    let cardDetails = {
      cardholderName: context.querySelector('[data-card-details="cardholder-name"]').value,
      cardNumber: cardNumber,
      expiryDate: context.querySelector('[data-card-details="expiry-date-month"]').value + context.querySelector('[data-card-details="expiry-date-year"]').value.substring(2),
      securityCode: context.querySelector('[data-card-details="security-code"]').value,
    };
    console.log('about to call sagepayOwnForm.tokeniseCardDetails with merchant session key: ' + settings.commerceopayopi.opayoCheckout.merchantSessionKey + ' for card: ' + cardNumber);
    sagepayOwnForm({ merchantSessionKey: settings.commerceopayopi.opayoCheckout.merchantSessionKey }).tokeniseCardDetails({
      cardDetails: cardDetails,
      onTokenised: function (result) { cardDetailsTokenised(result, settings, context); }
    });

  }


  function cardDetailsTokenised(result, settings, context) {
    console.log('sagepayOwnForm.tokeniseCardDetails result: ' + result.success + '(merchant session key ' + settings.commerceopayopi.opayoCheckout.merchantSessionKeySequenceNr + ': ' + settings.commerceopayopi.opayoCheckout.merchantSessionKey + ')');

    let checkoutForm = context.querySelector('form[id^="commerce-checkout"]');
    try {
      // Required field for form submit
      context.querySelector('#card-identifier-result').value = result.success;
    }
    catch { }
    if (result.success) {
      if (settings.commerceopayopi.opayoCheckout.mskTimeout) {
        clearTimeout(settings.commerceopayopi.opayoCheckout.mskTimeout);
      }
      context.querySelector('#card-identifier').value = result.cardIdentifier;
      // Return the merchant session key that was used for the tokenisation
      context.querySelector('#merchant-session-key').value = settings.commerceopayopi.opayoCheckout.merchantSessionKey;
      context.querySelector('#merchant-session-key-sequence-nr').value = settings.commerceopayopi.opayoCheckout.merchantSessionKeySequenceNr;
      context.querySelector('#merchant-session-key-expiry').value = settings.commerceopayopi.opayoCheckout.merchantSessionKeyExpiry;

      console.log('sagepayOwnForm.tokeniseCardDetails card identifier: ' + result.cardIdentifier + '(merchant session key ' + settings.commerceopayopi.opayoCheckout.merchantSessionKeySequenceNr + ': ' + settings.commerceopayopi.opayoCheckout.merchantSessionKey + ')');

      try {
        // Post the card identifier expiry back as well
        // Documentation states card identifier valid for 400 seconds
        // The exact value is not passed back by the Opayo javascript so assume 5 seconds for processing
        let cardIdExpTime = new Date();
        cardIdExpTime = new Date(cardIdExpTime.getTime() + (1000 * 395));
        let cardIdExpTimeField = context.querySelector('#card-identifier-expiry');
        cardIdExpTimeField.value = cardIdExpTime.toISOString();

        // Post back browser settings to be used with the (required) 'strong customer authentication' fields during the transaction request
        let colorDepth = window.screen.colorDepth;
        let screenHeight = window.innerHeight || document.documentElement.clientHeight || document.body.clientHeight;
        let screenWidth = window.innerWidth || document.documentElement.clientWidth || document.body.clientWidth;
        let timeZoneOffset = (new Date()).getTimezoneOffset();
        let language = navigator.language || navigator.userLanguage;
        // Create fields dynamically, haven't set up any preallocated hidden fields
        let input = document.createElement("input");
        input.setAttribute("type", "hidden");
        input.setAttribute("name", "browser-color-depth");
        input.setAttribute("value", colorDepth);
        checkoutForm.appendChild(input);
        input = document.createElement("input");
        input.setAttribute("type", "hidden");
        input.setAttribute("name", "browser-screen-height");
        input.setAttribute("value", screenHeight);
        checkoutForm.appendChild(input);
        input = document.createElement("input");
        input.setAttribute("type", "hidden");
        input.setAttribute("name", "browser-screen-width");
        input.setAttribute("value", screenWidth);
        checkoutForm.appendChild(input);
        input = document.createElement("input");
        input.setAttribute("type", "hidden");
        input.setAttribute("name", "browser-timezone-offset");
        input.setAttribute("value", timeZoneOffset);
        checkoutForm.appendChild(input);
        input = document.createElement("input");
        input.setAttribute("type", "hidden");
        input.setAttribute("name", "browser-language");
        input.setAttribute("value", language);
        checkoutForm.appendChild(input);
        // Card Type and last four digits
        input = document.createElement("input");
        input.setAttribute("type", "hidden");
        input.setAttribute("name", "card-type");
        input.setAttribute("value", settings.commerceopayopi.opayoCheckout.cardType);
        checkoutForm.appendChild(input);
        input = document.createElement("input");
        input.setAttribute("type", "hidden");
        input.setAttribute("name", "card-last-four-digits");
        input.setAttribute("value", settings.commerceopayopi.opayoCheckout.cardLastFourDigits);
        checkoutForm.appendChild(input);
      }
      catch { }

      checkoutForm.submit();
    }
    else {
      // Failed tokenisation - we still submit the form because we need a new merchant session key from the back-end
      try {
        context.querySelector('#card-identifier-error-httpcode').value = result.httpErrorCode;
        try {
          // If there are multiple errors, add the errors to the form as hidden fields
          for (let i = 0; i < result.errors.length; i++) {
            let errorDetail = result.errors[i];
            let input = document.createElement("input");
            input.setAttribute("type", "hidden");
            input.setAttribute("name", "card-identifier-error-" + i + "-code");
            input.setAttribute("value", errorDetail.code);
            checkoutForm.appendChild(input);
            input = document.createElement("input");
            input.setAttribute("type", "hidden");
            input.setAttribute("name", "card-identifier-error-" + i + "-message");
            input.setAttribute("value", errorDetail.message);
            checkoutForm.appendChild(input);
          }
        }
        catch { }
      }
      catch { }
      checkoutForm.submit();
    }
  }


  function startMerchantSessionKeyTimeout(msTimeout, merchantSessionKey, orderId) {
    // Start a timeout based on the merchant session key's validity
    let timeoutRef = setTimeout((merchantSessionKey, orderId) => {
      const spContainer = document.getElementById('sp-container');
      if (spContainer) {
        // Still displaying the checkout form, call the URL to refresh the merchant session key and return Ajax commands
        let ajaxObject = Drupal.ajax({
          url: '/commerce_opayo_pi/merchantsessionkey/' + orderId,
          base: false,
          element: false,
          progress: false
        });
        ajaxObject.execute();
      }
    }, msTimeout, merchantSessionKey, orderId);
    return timeoutRef;
  }

})(Drupal, once);
