<?php
namespace Drupal\commerce_opayo_pi\Controller;

use DateTime;
use Drupal\commerce_cart\CartProviderInterface;
use Drupal\commerce_cart\CartSessionInterface;
use Drupal\commerce_checkout\CheckoutOrderManagerInterface;
use Drupal\commerce_opayo_pi\Ajax\Opayo3DSecureResult;
use Drupal\Core\Controller\ControllerBase;
use Drupal\commerce_order\Entity\Order;
use Drupal\commerce_opayo_pi\Ajax\OpayoMerchantSessionKey;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Exception;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;

class OpayoController extends ControllerBase {

  /**
   * The checkout order manager.
   *
   * @var \Drupal\commerce_checkout\CheckoutOrderManagerInterface
   */
  protected $checkoutOrderManager;

  /**
   * The cart session.
   *
   * @var \Drupal\commerce_cart\CartSessionInterface
   */
  protected $cartSession;

  /**
   * The cart provider.
   *
   * @var \Drupal\commerce_cart\CartProviderInterface
   */
  protected $cartProvider;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The request stack
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * Logger
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * Constructs a new CheckoutController object.
   *
   * @param \Drupal\commerce_checkout\CheckoutOrderManagerInterface $checkout_order_manager
   *   The checkout order manager.
   * @param \Drupal\commerce_cart\CartSessionInterface $cart_session
   *   The cart session.
   * @param \Drupal\commerce_cart\CartProviderInterface $cart_provider
   *   The cart session.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger.
   * @param \Symfony\Component\HttpFoundation\RequestStack $requestStack
   *   The request stack.
   */
  public function __construct(CheckoutOrderManagerInterface $checkout_order_manager, CartSessionInterface $cart_session, CartProviderInterface $cart_provider, MessengerInterface $messenger, EntityTypeManagerInterface $entity_type_manager, RequestStack $requestStack, LoggerChannelInterface $logger) {
    $this->checkoutOrderManager = $checkout_order_manager;
    $this->cartSession = $cart_session;
    $this->cartProvider = $cart_provider;
    $this->messenger = $messenger;
    $this->entityTypeManager = $entity_type_manager;
    $this->requestStack = $requestStack;
    $this->logger = $logger;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('commerce_checkout.checkout_order_manager'),
      $container->get('commerce_cart.cart_session'),
      $container->get('commerce_cart.cart_provider'),
      $container->get('messenger'),
      $container->get('entity_type.manager'),
      $container->get('request_stack'),
      $container->get('commerce_opayo_pi.logger.channel.commerce_opayo_pi'),
    );
  }


  /**
   * During 3DS authentication the ACS redirects the cardholder back to the URL represented by 'Process3DSecure' (all within the context of the
   * iFrame set up within the 'Opayo3DSReview' checkout pane)
   *
   * @param \Drupal\Core\Routing\RouteMatchInterface $route_match
   *   The route match.
   *
   */
  public function Process3DSecure(RouteMatchInterface $route_match) {

    $request = $this->requestStack->getCurrentRequest();
    try {
      $this->logger->info('OpayoController::Process3DSecure: from {clientIp}', ['clientIp' => $request->getClientIp()]);
    } catch (Exception $le) {
    }

    // We provided the Opayo 'transaction id' as the 'threeDSSessionData' when starting the 3DS challenge - note the 'transaction_id' is still BASE64 encoded
    /** @var \Drupal\commerce_order\Entity\OrderInterface $order */
    $order = $route_match->getParameter('order');

    // Use symphony to get the parameters from the POST
    $transaction_id_base64 = $request->request->get('threeDSSessionData');
    $challenge_response = $request->request->get('cres');
    $payment_gateway_plugin = null;

    // We need the configuration info from the payment gateway
    // Opayo transaction ID is linked to the Order ID, which can be used to instantiate an Opayo payment gateway object with its config data
    if ($transaction_id_base64)
    {
      $opayo_transaction_id = base64_decode($transaction_id_base64);
      // Find the Opayo transaction
      /** @var \Drupal\commerce_opayo_pi\Entity\OpayoTransaction $opayo_transaction */
      $opayo_transaction = null;
      try {
        $query = $this->entityTypeManager->getStorage('opayo_transaction')->getQuery();
        $query->condition('transaction_id', $opayo_transaction_id);
        $query->condition('c_req', null, 'IS NOT NULL');
        $query->sort('received', 'DESC');
        $ids = $query->execute();
        if (sizeof($ids) >= 1) {
          /** @var \Drupal\commerce_opayo_pi\Entity\OpayoTransaction $opayo_transaction */
          $opayo_transaction = $this->entityTypeManager->getStorage('opayo_transaction')->load(array_values($ids)[0]);
        }
      } catch (Exception $exception) {
      }
      if ($opayo_transaction != null) {
        try {
          /** @var \Drupal\commerce_payment\PaymentGatewayStorageInterface $payment_gateway_storage */
          $payment_gateway_storage = $this->entityTypeManager->getStorage('commerce_payment_gateway');
          $payment_gateways = $payment_gateway_storage->loadMultipleForOrder($order);
          $payment_gateway = $payment_gateways['opayo_pi'];
          $payment_gateway_plugin = $payment_gateway->getPlugin();

          // Update the opayo transaction state to indicate that we've received a challenge response
          $opayo_transaction->set('c_res', $challenge_response);
          $time_now = new DateTime();
          $opayo_transaction->set('3d_response_received', $time_now->getTimestamp());
          $opayo_transaction->save();

          /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $payment_gateway_plugin  */
          if ($payment_gateway_plugin != null && $order != null)
          {
            try {
              // Opayo Pi Payment gateway to use the 3DS response challenge result to request payment transaction
              $result = $payment_gateway_plugin->process3DSAuthenticationResult($order, $transaction_id_base64, $opayo_transaction, $challenge_response);

              if ($result) {
                try {
                  $this->logger->info('OpayoController::Process3DSecure: successful payment transaction after successful 3D secure challenge response for transaction {transactionId}, (base64: {base64TransId}), client IP: {clientIp}',
                    ['transactionId' => $opayo_transaction_id, 'base64TransId' => $transaction_id_base64, 'clientIp' => $request->getClientIp(), 'clientIp' => $request->getClientIp()]);
                } catch (Exception $le) {
                }

                // Return a page (within the iFrame) that sends a success message to the iFrame's parent
                // Render array
                return [
                  '#markup' => '<p>3D Secure success</p>',
                  '#prefix' => '<div id="commerce-opayo-3d-success">',
                  '#suffix' => '</div>',
                  '#attached' => [
                    'library' => ['commerce_opayo_pi/opayo_pi_3dsecure_iframe'],
                  ],
                ];
              }
            }
            catch (Exception $pe) {
              try {
                $this->logger->error('OpayoController::Process3DSecure: internal error processing transaction {transactionId}, (base64: {base64TransId}), internal exception: {exception}',
                  ['transactionId' => $opayo_transaction_id, 'base64TransId' => $transaction_id_base64, 'clientIp' => $request->getClientIp(), 'exception' => ($pe->getMessage() . ' ' . $pe->getTraceAsString())]);
              } catch (Exception $le) {
              }
            }
          }
          else {
            try {
              $this->logger->error('OpayoController::Process3DSecure: internal error processing transaction {transactionId}, (base64: {base64TransId})', ['transactionId' => $opayo_transaction_id, 'base64TransId' => $transaction_id_base64, 'clientIp' => $request->getClientIp()]);
            } catch (Exception $le) {
            }
          }
        }
        catch (Exception $e) {
          $this->logger->error('OpayoController::Process3DSecure: exception processing transaction {transactionId}, (base64: {base64TransId}), exception: {exc}', ['exc' => ($e->getMessage() . ' ' . $e->getTraceAsString()), 'transactionId' => $opayo_transaction_id, 'base64TransId' => $transaction_id_base64, 'clientIp' => $request->getClientIp()]);
        }
      }
      else {
        // No matching transaction found
        try {
          $this->logger->warning('OpayoController::Process3DSecure: failed to find transaction {transactionId}, (base64: {base64TransId})', ['transactionId' => $opayo_transaction_id, 'base64TransId' => $transaction_id_base64, 'clientIp' => $request->getClientIp()]);
        } catch (Exception $le) {
        }
      }
    }
    else {
      try {
        $this->logger->warning('OpayoController::Process3DSecure: request from from {clientIp} missing threeDSSessionData, payload: {payload}', ['payload' => $request->getContent(), 'clientIp' => $request->getClientIp()]);
      } catch (Exception $le) {
      }
    }


    // User feedback needs to be passed from within the iFrame to the parent
    $computed_settings = [
      'message' => '3DS authentication failure. You need to re-enter your card details and retry.',
    ];

    // Return a page (within the iFrame) that sends a failure message to the iFrame's parent
    return [
      '#markup' => '<p>3D Secure failure</p>',
      '#prefix' => '<div id="commerce-opayo-3d-failure">',
      '#suffix' => '</div>',
      '#attached' => [
        'library' => ['commerce_opayo_pi/opayo_pi_3dsecure_iframe'],
        'drupalSettings' => ['commerceopayopi' => ['opayo_3dsecure' => $computed_settings]],
      ],
    ];

  }


  /**
   * Returns the status/result of 3DS authentication in an AJAX javascript response for a given order and opayo transaction ID
   *
   * @param \Drupal\Core\Routing\RouteMatchInterface $route_match
   *   The route match.
   *
   */
  public function Get3DSecureResult(RouteMatchInterface $route_match) {

    /** @var \Drupal\commerce_order\Entity\OrderInterface $order */
    $order = $route_match->getParameter('order');

    /** @var \Drupal\commerce_opayo_pi\Entity\OpayoTransaction $opayo_transaction */
    $opayo_transaction = $route_match->getParameter('opayo_transaction_id');

    $result = 'transaction_not_found';

    if ($opayo_transaction != null)
    {
      if ($opayo_transaction->getStatusCode() == '0000' && $opayo_transaction->getTransactionType() == 'Payment') {
        $result = 'success';
      }
      elseif ($opayo_transaction->getStatus() == "3DAuth") {
        // 3D Secure authentication started
        /** @var int $threed_response_received_time */
        $threed_response_received_time = $opayo_transaction->get('3d_response_received')->value;

        if ($threed_response_received_time > 0)
        {
          // A 3D response has been received and processed, check the result
          $status_code = $opayo_transaction->getStatusCode();
          if ($status_code == '2000') {
            $result = 'declined';
          }
          else {
            $result = 'failed';
          }
        }
        else {
          // No response yet from the 3D secure process, check expirty
          $received_time = $opayo_transaction->getReceived();
          $seconds_since_received = (new DateTime('now'))->getTimestamp() - $received_time;
          if ($seconds_since_received > 1200)
            $result = 'expired';
          else
            $result = 'inprogress';
        }
      }
    }

    $response = new AjaxResponse();

    $computed_settings = [
      'transactionId' => $opayo_transaction->getTransactionId(),
      'orderId' => $order->id(),
      'result' => $result,
    ];
    if ($result == 'declined' || $result == 'failed')
    {
      $computed_settings['message'] = '3DS authentication failure. You need to re-enter your card details and retry.';
    }
    elseif ($result == 'expired')
    {
      $computed_settings['message'] = '3DS authentication not completed within 20 minutes. You need to re-enter your card details and retry.';
    }

    // Add an AJAX javascript command to process an new merchant session key in the front-end javascript
    $response->addCommand(new Opayo3DSecureResult($computed_settings));

    return $response;


  }

  /**
   * Requests and returns a new merchant session key in an AJAX response.
   *
   * @param \Drupal\Core\Routing\RouteMatchInterface $route_match
   *   The route match.
   *
   * @return array|Drupal\Core\Ajax\AjaxResponse
   *   The Ajax response commands
   */
  public function MerchantSessionKey(RouteMatchInterface $route_match) {

    $order = $route_match->getParameter('commerce_order');
    if ($order != null)
    {
      // The request may have an 'instruction' that gets returned with the Ajax response alongside the new merchant session key, to be
      // acted upon by the custom 'opayoMerchantSessionKey' Ajax command
      $instruction = $route_match->getParameter('instruction');

      // Get the payment gateway plugin via the order through the following scaffolding:
      /** @var \Drupal\commerce_payment\PaymentGatewayStorageInterface $payment_gateway_storage */
      $payment_gateway_storage = $this->entityTypeManager->getStorage('commerce_payment_gateway');
      $payment_gateways = $payment_gateway_storage->loadMultipleForOrder($order);
      $payment_gateway = $payment_gateways['opayo_pi'];
      /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $payment_gateway_plugin  */
      $payment_gateway_plugin = $payment_gateway->getPlugin();

      $response = new AjaxResponse();

      /** @var StdClass $newMerchantSessionKey  */
      $newMerchantSessionKey = $payment_gateway_plugin->getNewMerchantSessionKey($order->id());
      $expiryinsecs = (new DateTime($newMerchantSessionKey->expiry))->getTimestamp() - (new DateTime('now'))->getTimestamp();
      $computed_settings = [
        'merchantSessionKey' => $newMerchantSessionKey->merchantSessionKey,
        'merchantSessionKeyExpirySecs' => $expiryinsecs,
        'merchantSessionKeyExpiry' => $newMerchantSessionKey->expiry,
        'merchantSessionKeySequenceNr'=> $newMerchantSessionKey->sequenceNr,
        'orderId' => $order->id(),
      ];

      if (isset($instruction))
        // Pass the instruction back to the javascript
        $computed_settings['instruction'] = $instruction;

      // Add our custom 'opayoMerchantSessionKey' AJAX javascript command to process an new merchant session key in the front-end javascript
      $response->addCommand(new OpayoMerchantSessionKey($computed_settings));

      return $response;
    }
  }

}
