<?php

namespace Drupal\commerce_opayo_pi\PluginForm\OpayoPi;

use Drupal\commerce_payment\Exception\DeclineException;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\PluginForm\PaymentMethodAddForm as BasePaymentMethodAddForm;
use Drupal\commerce_payment\PluginForm\PaymentMethodFormBase;
use Drupal\Core\Form\FormStateInterface;
use DateTime;
use Drupal;
use Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGateway;
use Exception;
use stdClass;

class PaymentMethodAddForm extends BasePaymentMethodAddForm {

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {

    $form = PaymentMethodFormBase::buildConfigurationForm($form, $form_state);

    /** @var \Drupal\commerce_payment\Entity\PaymentMethodInterface $payment_method */
    $payment_method = $this->entity;

    $form['payment_details'] = [
      '#parents' => array_merge($form['#parents'], ['payment_details']),
      '#type' => 'container',
      '#payment_method_type' => $payment_method->bundle(),
    ];

    if ($payment_method->bundle() == 'credit_card_opayo') {
      /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin */
      $pgplugin = $this->plugin;
      $gateway_config = $pgplugin->getConfiguration();
      if ($gateway_config['form_type'] == 'dropin') {
        // Use drop-in version of the form
        $form['payment_details'] = $this->buildOpayoPiDropinForm($form['payment_details'], $form_state);
      }
      else {
        // Use the 'own form' version
        $form['payment_details'] = $this->buildOpayoPiOwnForm($form['payment_details'], $form_state);
        // Use an #after_build directive to tweak the payment details fields
        $form['payment_details']['#after_build'][] = [get_class($this), 'customisePaymentDetails'];
      }
    }

    // Use an #after_build directive to tweak the billing address fields
    if ($gateway_config['form_type'] == 'ownform') {
      // For 'own form', have cardholder first name and last name with the card details fields rather than with the address fields
      $form['billing_information']['address']['widget'][0]['address']['#after_build'][] = [get_class($this), 'customiseAddress'];
    }

    // Use an #after_build directive to tweak the billing information fields
    $form['billing_information']['#after_build'][] = [get_class($this), 'customiseBillingInfo'];


    // Move the billing information below the payment details.
    if (isset($form['billing_information'])) {
      $form['billing_information']['#weight'] = 10;
    }

    return $form;
  }

  /**
   * Tweak some aspects of the address fields for 'own form'
   */
  public static function customisePaymentDetails(array $element, FormStateInterface $form_state) {

    // We want the credit card field form elements to be displayed (and filled in by the user) but not get returned with the post (a tokenised 'card identifier' is returned instead)
    // Taking away the 'name' attribute should ensure this
    unset($element['cardholder-name']['#name']);
    unset($element['number']['#name']);
    unset($element['expiration']['month']['#name']);
    unset($element['expiration']['year']['#name']);
    unset($element['security-code']['#name']);
    // Opayo require the 'first name' and 'last name' when submitting the transaction so these fields are not excluded from the 'post'

    return $element;
  }

  /**
   * Tweak some aspects of the address fields
   */
  public static function customiseAddress(array $element, FormStateInterface $form_state) {

    // We already have the cardholder name field (both for 'drop-in' and 'own form') so don't need the first name, middle name and surname as well
    unset($element['container0']);
    unset($element['given_name']);
    unset($element['family_name']);
    unset($element['additional_name']);

    return $element;
  }

  /**
   * Tweak some aspects of the billing information
   */
  public static function customiseBillingInfo(array $element, FormStateInterface $form_state) {

    if ($form_state->getValue('process_input') == false) {
      // Match the phone number's country to the billing country

      // This is only effective on initial form build. When the user changes the country the Javascript only affects the 'address' section and
      // will not update the international telephone number field. Update in javascript instead.
      $country_code = null;
      try {
        if (array_key_exists('address', $element) && array_key_exists('widget', $element['address']) && array_key_exists(0, $element['address']['widget']) && array_key_exists('address', $element['address']['widget'][0]) && array_key_exists('country_code', $element['address']['widget'][0]['address'])) {
          $country_code = $element['address']['widget'][0]['address']['country_code']['#value'];
          if ($country_code != null) {
            // Need to get the field name for the telephone number from the gateway configuration
            $payment_gateway_storage = \Drupal::entityTypeManager()->getStorage('commerce_payment_gateway');
            $payment_gateways = $payment_gateway_storage->loadByProperties(['status' => TRUE]);
            /** @var \Drupal\commerce_payment\Entity\PaymentGatewayInterface $payment_gateway */
            $payment_gateway = $payment_gateways['opayo_pi'];
            /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $plugin  */
            $plugin = $payment_gateway->getPlugin();
            $configuration = $plugin->getConfiguration();
            $telephoneFieldName = $configuration['telephone_field_name'];

            try {
              // If the telephone field is of type 'phone_international' (https://www.drupal.org/project/phone_international) then the following code copies the country code
              $element[$telephoneFieldName]['widget'][0]['value']['#country'] = $country_code;
              $element[$telephoneFieldName]['widget'][0]['value']['int_phone']['#attributes']['data-country'] = $country_code;
            }
            catch (Exception $ite) {}
          }
        }
      }
      catch (Exception $e) {
      }
    }

    return $element;
  }


  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    PaymentMethodFormBase::validateConfigurationForm($form, $form_state);
    /** @var \Drupal\commerce_payment\Entity\PaymentMethodInterface $payment_method */
    $payment_method = $this->entity;

    if ($payment_method->bundle() == 'credit_card_opayo') {
      /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin */
      $pgplugin = $this->plugin;
      $gateway_config = $pgplugin->getConfiguration();
      if ($gateway_config['form_type'] == 'dropin') {
        // Use drop-in version of the form
        $this->validateOpayoPiDropinForm($form['payment_details'], $form_state);
      }
      else {
        $this->validateOpayoPiOwnForm($form['payment_details'], $form_state);

        if ($form_state->isSubmitted())
        {
          // Need to copy the first name and last name from the 'payment details' section to the billing profile
          // (if it's done here the changes will get saved to the database - cannot wait until 'submitConfigurationForm')
          $cardHolderFirstname = $form_state->getValue($form["payment_details"]["cardholder"]["cardholder-first-name"]['#parents']);
          $cardHolderLastname = $form_state->getValue($form["payment_details"]["cardholder"]["cardholder-last-name"]['#parents']);
          $billingAddress = $form_state->getValue($form["billing_information"]["address"]["widget"][0]["address"]["#parents"]);
          $billingAddress['given_name'] = $cardHolderFirstname;
          $billingAddress['family_name'] = $cardHolderLastname;
          $form_state->setValueForElement($form["billing_information"]["address"]["widget"][0]["address"], $billingAddress);
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    PaymentMethodFormBase::submitConfigurationForm($form, $form_state);
    /** @var \Drupal\commerce_payment\Entity\PaymentMethodInterface $payment_method */
    $payment_method = $this->entity;
    /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin */
    $pgplugin = $this->plugin;

    if ($payment_method->bundle() == 'credit_card_opayo') {
      /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin */
      $pgplugin = $this->plugin;
      $gateway_config = $pgplugin->getConfiguration();
      if ($gateway_config['form_type'] == 'dropin') {
        // Use drop-in version of the form
        $this->submitOpayoPiDropinForm($form['payment_details'], $form_state);
      }
      else {
        // 'Own Form' version
        $this->submitOpayoPiOwnForm($form['payment_details'], $form_state);
      }
    }

    // Submitted checkout form values
    $form_values = $form_state->getValue($form['#parents']);

    // Add the order id
    $order = $pgplugin->getOrder();
    $form_values['payment_details']['order-id'] = $order->id();

    if ($gateway_config['form_type'] == 'ownform') {
      // Card type and last four digits supplied by javascript
      $user_input = $form_state->getUserInput();
      try {
        if (array_key_exists('card-type', $user_input))
          $form_values['payment_details']['card-type'] = $user_input['card-type'];
        if (array_key_exists('card-last-four-digits', $user_input))
          $form_values['payment_details']['card-last-four-digits'] = $user_input['card-last-four-digits'];
      } catch (Exception $e) {
      }
    }
    else {
      // Dropin form - add merchant session key info to the payment details before calling the plugin
      $user_input = $form_state->getUserInput();
      $form_values['payment_details']['merchant-session-key'] = $user_input["payment_information"]["add_payment_method"]["payment_details"]["merchant-session-key"];
      $form_values['payment_details']['merchant-session-key-sequence-nr'] = $user_input["payment_information"]["add_payment_method"]["payment_details"]["merchant-session-key-sequence-nr"];
      $form_values['payment_details']['merchant-session-key-expiry'] = $user_input["payment_information"]["add_payment_method"]["payment_details"]["merchant-session-key-expiry"];
    }

    // The payment method form is customer facing. For security reasons
    // the returned errors need to be more generic.
    try {
      $pgplugin->createPaymentMethod($payment_method, $form_values['payment_details']);
    }
    catch (DeclineException $e) {
      $this->logger->warning($e->getMessage());
      throw new DeclineException(t('We encountered an error processing your payment method. Please verify your details and try again.'));
    }
    catch (PaymentGatewayException $e) {
      $this->logger->error($e->getMessage());
      throw new PaymentGatewayException(t('We encountered an unexpected error processing your payment method. Please try again later.'));
    }
  }


  /**
   * Builds the credit card form.
   *
   * @param array $element
   *   The target element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   *
   * @return array
   *   The built credit card form.
   */
  protected function buildOpayoPiDropinForm(array $element, FormStateInterface $form_state) {

    /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin  */
    $pgplugin = $this->plugin;
    $order = $pgplugin->getOrder();

    // Merchant session key needs to be presupplied

    /** @var StdClass $merchant_session_key  */
    $merchant_session_key = $pgplugin->getNewMerchantSessionKey($order->id());

    // How many seconds until expiry
    $expiryinsecs = (new DateTime($merchant_session_key->expiry))->getTimestamp() - (new DateTime('now'))->getTimestamp();

    $computed_settings = [
      'merchantSessionKey' => $merchant_session_key->merchantSessionKey,
      'merchantSessionKeyExpirySecs' => $expiryinsecs,
      'merchantSessionKeyExpiry' => $merchant_session_key->expiry,
      'merchantSessionKeySequenceNr' => $merchant_session_key->sequenceNr,
      'orderId' => $order->id(),
    ];

    // Javascript that calls Opayo's 'sagepayCheckout' function through Drupal.behaviors
    $element['#attached']['library'][] = 'commerce_opayo_pi/opayo_pi_dropinform';
    // Pass the settings including the merchant session to the Javascript
    $element['#attached']['drupalSettings']['commerceopayopi']['opayoCheckout'] = $computed_settings;

    $gateway_config = $pgplugin->getConfiguration();
    $opayo_js_ref = null;
    if($gateway_config['mode'] == 'test')
      $opayo_js_ref = 'commerce_opayo_pi/opayo_sagepay_test';
    else
       $opayo_js_ref = 'commerce_opayo_pi/opayo_sagepay_live';
    // Attach the external Opayo provided javascript
    $element['#attached']['library'][] = $opayo_js_ref;

    // Feedback about expiry of the card fields
    $element['expiry-msg'] = [
      '#type' => 'markup',
      '#markup' => '<div id="expiry-msg" class="commerce-opayo-pi-expiry" style="font-size:smaller"></div>',
    ];

    // Hidden elements with the merchant session key info
    $element['merchant-session-key'] = [
      '#type' => 'hidden',
      '#value' => $merchant_session_key->merchantSessionKey,
      '#attributes' => [
        'id' => ['merchant-session-key'],
      ],
    ];
    $element['merchant-session-key-sequence-nr'] = [
      '#type' => 'hidden',
      '#value' => $merchant_session_key->sequenceNr,
      '#attributes' => [
        'id' => ['merchant-session-key-sequence-nr'],
      ],
    ];
    $element['merchant-session-key-expiry'] = [
      '#type' => 'hidden',
      '#value' => $merchant_session_key->expiry,
      '#attributes' => [
        'id' => ['merchant-session-key-expiry'],
      ],
    ];

    // HTML element that will contain the iFrame for Opayo's 'Drop-In' elements
    $element['sp-container'] = [
      '#type' => 'container',
      '#attributes' => [
        'id' => ['sp-container'],
      ],
    ];

    return $element;
  }


  /**
   * Validates the credit card form.
   *
   * @param array $element
   *   The credit card form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   */
  protected function validateOpayoPiDropinForm(array &$element, FormStateInterface $form_state) {

    if ($form_state->isSubmitted())
    {
      $cardIdentifier = '';
      $cardIdentifierHttpCode = 0;
      $cardIdentifierErrorCode = 0;
      $cardIdentifierErrorMessage = '';

      if (array_key_exists('card-identifier', $form_state->getUserInput()))
        // Opayo's 'Drop-In' logic should have provided a 'Card Identifier' among the posted fields
        $cardIdentifier = $form_state->getUserInput()['card-identifier'];

      if (!$cardIdentifier) {
        // Check to see if we had an expired merchant session key

        /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGateway $pgplugin  */
        $pgplugin = $this->plugin;
        $currentMerchantSessionKey = $pgplugin->getCurrentMerchantSessionKey($pgplugin->getOrder()->id());
        if ($currentMerchantSessionKey != null)
        {
          if (new DateTime('now') >= DateTime::createFromFormat('U', $currentMerchantSessionKey->expiry))
            $form_state->setErrorByName('sp-container', $this->t('Your card details could not be validated.'));
        }

        // Check for any error fields in the POST that were supplied by the Opayo javascript
        if (array_key_exists('card-identifier-http-code', $form_state->getUserInput()))
          $cardIdentifierHttpCode = $form_state->getUserInput()['card-identifier-http-code'];
        if (array_key_exists('card-identifier-error-code', $form_state->getUserInput()))
          $cardIdentifierErrorCode = $form_state->getUserInput()['card-identifier-error-code'];
        if (array_key_exists('card-identifier-error-message', $form_state->getUserInput()))
          $cardIdentifierErrorMessage = $form_state->getUserInput()['card-identifier-error-message'];

        if (isset($cardIdentifierHttpCode) && isset($cardIdentifierErrorCode) && isset($cardIdentifierErrorMessage))
        {
          // TODO: refine
          $form_state->setErrorByName('sp-container', 'HTTP Code: ' . $cardIdentifierHttpCode . ', Error code: ' . $cardIdentifierErrorCode . ', Error message: ' . $cardIdentifierErrorMessage);
        }
        else {
          // Display a generic error
          $form_state->setErrorByName('sp-container', $this->t('Your card details could not be validated.'));
        }

        // Invalidate the current merchant session key
        /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGateway $pgplugin  */
        $pgplugin = $this->plugin;
        $pgplugin->invalidateCurrentMerchantSessionKey($pgplugin->getOrder()->id());
        return;
      }

      // Do not call the parent method (it would try to validate card details that are handled by the drop-in form instead)
    }
  }

  /**
   * Handles the submission of the credit card form.
   *
   * @param array $element
   *   The credit card form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   */
  protected function submitOpayoPiDropinForm(array $element, FormStateInterface &$form_state) {
    // Opayo's 'Drop-In' logic should have provided a 'Card Identifier' among the posted fields
    $card_identifier = $form_state->getUserInput()['card-identifier'];

    // Update the form_state to contain the payment details, in this case the only payment detail is the 'card_identifier' token
    $payment_details = [ 'card-identifier' => $card_identifier ];
    $form_state->setValueForElement($element, $payment_details);

    // Current merchant session key has been used succesfully  with the card details to generate a 'card identifier'. It cannot be
    // used again, invalidate it
    /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGateway $pgplugin  */
    $pgplugin = $this->plugin;
    $pgplugin->invalidateCurrentMerchantSessionKey($pgplugin->getOrder()->id());

  }

  /**
   * Builds the credit card form.
   *
   * @param array $element
   *   The target element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   *
   * @return array
   *   The built credit card form.
   */
  protected function buildOpayoPiOwnForm(array $element, FormStateInterface $form_state) {

    // Credit card fields
    $element = $this->buildCreditCardForm($element, $form_state);

    /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin  */
    $pgplugin = $this->plugin;
    $order = $pgplugin->getOrder();

    // Pass through the order ID
    $computed_settings = [
      'orderId' => $order->id(),
    ];

    // Javascript that calls 'sagepayOwnForm' through Drupal.behaviors
    $element['#attached']['library'][] = 'commerce_opayo_pi/opayo_pi_ownform';
    // Pass the settings including the merchant session to the Javascript
    $element['#attached']['drupalSettings']['commerceopayopi']['opayoCheckout'] = $computed_settings;

    $gateway_config = $pgplugin->getConfiguration();
    $opayo_js_ref = null;
    if ($gateway_config['mode'] == 'test')
    $opayo_js_ref = 'commerce_opayo_pi/opayo_sagepay_test';
    else
      $opayo_js_ref = 'commerce_opayo_pi/opayo_sagepay_live';
    // Attach the external Opayo provided javascript
    $element['#attached']['library'][] = $opayo_js_ref;

    // hidden element for the card identifier pass-back
    $element['card-identifier'] = [
      '#type' => 'hidden',
      '#value' => '',
      '#attributes' => [
        'id' => ['card-identifier'],
      ],
    ];
    $element['card-identifier-expiry'] = [
      '#type' => 'hidden',
      '#value' => '',
      '#attributes' => [
        'id' => ['card-identifier-expiry'],
      ],
    ];
    $element['card-identifier-result'] = [
      '#type' => 'hidden',
      '#value' => '',
      '#attributes' => [
        'id' => ['card-identifier-result'],
      ],
      '#required' => TRUE,
    ];
    $element['card-identifier-error-httpcode'] = [
      '#type' => 'hidden',
      '#value' => '',
      '#attributes' => [
        'id' => ['card-identifier-error-httpcode'],
      ],
    ];
    // Hidden elements with the merchant session key info
    $element['merchant-session-key'] = [
      '#type' => 'hidden',
      '#value' => '',
      '#attributes' => [
        'id' => ['merchant-session-key'],
      ],
    ];
    $element['merchant-session-key-sequence-nr'] = [
      '#type' => 'hidden',
      '#value' => 0,
      '#attributes' => [
        'id' => ['merchant-session-key-sequence-nr'],
      ],
    ];
    $element['merchant-session-key-expiry'] = [
      '#type' => 'hidden',
      '#value' => '',
      '#attributes' => [
        'id' => ['merchant-session-key-expiry'],
      ],
    ];

    return $element;
  }

  /**
   * Builds the credit card form.
   *
   * @param array $element
   *   The target element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   *
   * @return array
   *   The built credit card form.
   */
  protected function buildCreditCardForm(array $element, FormStateInterface $form_state) {

    /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin */
      $pgplugin = $this->plugin;
      $gateway_config = $pgplugin->getConfiguration();

    // Build a month select list that shows months with a leading zero.
    $months = [];
    for ($i = 1; $i < 13; $i++) {
      $month = str_pad($i, 2, '0', STR_PAD_LEFT);
      $months[$month] = $month;
    }
    // Build a year select list that uses a 4 digit key with a 2 digit value.
    $current_year_4 = date('Y');
    $current_year_2 = date('y');
    $years = [];
    for ($i = 0; $i < 10; $i++) {
      $years[$current_year_4 + $i] = $current_year_2 + $i;
    }

    $element['#attributes']['class'][] = 'credit-card-form';
    // Placeholder for the detected card type. Set by validateCreditCardForm().
    $element['type'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];
    // Put cardholder first and last name with the card details
    $element['cardholder'] = [
      '#type' => 'container',
      '#attributes' => [
        'class' => ['credit-card-form__cardholder'],
      ],
    ];
    $element['cardholder']['cardholder-first-name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('First Name'),
      '#attributes' => ['autocomplete' => 'off', 'data-card-details' => 'cardholder-first-name'],
      '#required' => TRUE,
      '#maxlength' => 35,
      '#size' => 12,
    ];
    // Allow the surname to stay blank in 'test' mode so that the first name can be used for Opayo 'Magic CardHolder Values'
    $element['cardholder']['cardholder-last-name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Last Name'),
      '#attributes' => ['autocomplete' => 'off', 'data-card-details' => 'cardholder-last-name'],
      '#required' => TRUE,
      '#maxlength' => 35,
      '#size' => 12,
    ];
    $element['cardholder-name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Card Holder'),
      '#attributes' => ['autocomplete' => 'off', 'data-card-details' => 'cardholder-name'],
      '#required' => TRUE,
      '#maxlength' => 35,
      '#size' => 25,
    ];
    $element['number'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Card number'),
      '#attributes' => ['autocomplete' =>'off', 'data-card-details' => 'card-number'],
      '#required' => TRUE,
      '#maxlength' => 19,
      '#size' => 25,
    ];
    $element['expiration'] = [
      '#type' => 'container',
      '#attributes' => [
        'class' => ['credit-card-form__expiration'],
      ],
    ];
    $element['expiration']['month'] = [
      '#type' => 'select',
      '#title' => $this->t('Expiry Month'),
      '#attributes' => ['data-card-details' => 'expiry-date-month'],
      '#options' => $months,
      '#default_value' => date('m'),
      '#required' => TRUE,
    ];
    $element['expiration']['year'] = [
      '#type' => 'select',
      '#title' => $this->t('Expiry Year'),
      '#attributes' => ['data-card-details' => 'expiry-date-year'],
      '#options' => $years,
      '#default_value' => $current_year_4,
      '#required' => TRUE,
    ];
    $element['security-code'] = [
      '#type' => 'textfield',
      '#title' => $this->t('CVV'),
      '#attributes' => ['autocomplete' => 'off', 'data-card-details' => 'security-code'],
      '#required' => TRUE,
      '#maxlength' => 4,
      '#size' => 4,
    ];

    return $element;
  }

  /**
   * Validates the credit card form when using 'Own Form' appreach.
   *
   * @param array $element
   *   The credit card form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   */
  protected function validateOpayoPiOwnForm(array &$element, FormStateInterface $form_state) {

    if ($form_state->isSubmitted())
    {
      // Checks needed when payment details have been submitted

      /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin  */
      $pgplugin = $this->plugin;

      $card_identifier = null;
      $card_identifier_http_error_code = null;
      $card_identifier_errors = array();

      // The values, including 'payment_details', that are added to the form by the javascript are not found in the $form_state's 'values' but
      // can be retrieved from its 'input' field.

      $user_input = $form_state->getUserInput();
      try {
        $result = $user_input["payment_information"]["add_payment_method"]["payment_details"]["card-identifier-result"];
        $merchantSessionKey = $user_input["payment_information"]["add_payment_method"]["payment_details"]["merchant-session-key"];
        if ($result == "true")
        {
          $card_identifier = $user_input["payment_information"]["add_payment_method"]["payment_details"]["card-identifier"];
        }
        else {
          $card_identifier_http_error_code = $user_input["payment_information"]["add_payment_method"]["payment_details"]["card-identifier-error-httpcode"];

          $index = 0;
          if (array_key_exists("card-identifier-error-" . $index . "-code", $user_input))
          {
            $errorDetails = new stdClass();
            $errorDetails->code = $user_input["card-identifier-error-" . $index . "-code"];
            $errorDetails->message = $user_input["card-identifier-error-" . $index . "-message"];
            array_push($card_identifier_errors, $errorDetails);
            $index++;
          }
        }
      }
      catch (Exception $e) {
      }

      if ($form_state->hasAnyErrors()) {
        // Some error messages are the result of the fact that the credit card fields are on the form but their values are tokenised using javascript
        // and not returned here. Filter those out.
        $form_errors = $form_state->getErrors();
        $filtered_form_errors = array();
        /** @var Drupal\Core\StringTranslation\TranslatableMarkup $errorElement */
        foreach ($form_errors as $errorElement => $errorObject) {
          $ignore = (strval($errorElement) == 'payment_information][add_payment_method][payment_details][cardholder][cardholder-first-name' ||
            strval($errorElement) == 'payment_information][add_payment_method][payment_details][cardholder][cardholder-last-name' ||
            strval($errorElement) == 'payment_information][add_payment_method][payment_details][cardholder-name' ||
            strval($errorElement) == 'payment_information][add_payment_method][payment_details][number' ||
            strval($errorElement) == 'payment_information][add_payment_method][payment_details][security-code');
          if (!$ignore) {
            $filtered_form_errors[$errorElement] = $errorObject;
          }
        }
        if (sizeof($filtered_form_errors) < sizeof($form_errors)) {
          // Some errors have been filtered out - reset
          $form_state->clearErrors();
          foreach ($filtered_form_errors as $errorElement => $errorObject) {
            $form_state->setErrorByName($errorElement, $errorObject);
          }
        }
      }

      if ($card_identifier == null) {
        // No card-identifier returned with the form

        $date_utc = new \DateTime("now", new \DateTimeZone("UTC"));
        if ($card_identifier_http_error_code != null) {
          $errorFeedback = false;
          if ($card_identifier_http_error_code == 422)
          {
            $cardholder_name_warning = '';
            $card_number_warning = '';
            $card_expiry_warning = '';
            $card_cvv_warning = '';
            $token_problem = '';
            foreach($card_identifier_errors as $errorDetails) {
              if (str_contains(strtolower($errorDetails->message), 'card number')) {
                $card_number_warning .= (strlen($card_number_warning) > 0 ? ', ' : '') . $errorDetails->message;
              }
              else if (str_contains(strtolower($errorDetails->message), 'cardholder name')) {
                $cardholder_name_warning .= (strlen($cardholder_name_warning) > 0 ? ', ' : '') . $errorDetails->message;
              }
              else if (str_contains(strtolower($errorDetails->message), 'security code')) {
                $card_cvv_warning .= (strlen($card_cvv_warning) > 0 ? ', ' : '') . $errorDetails->message;
              }
              else if ($errorDetails->code == '1006' || $errorDetails->code == '1011')
                $token_problem .= (strlen($token_problem) > 0 ? ', ' : '') . $errorDetails->code;
              else if ($errorDetails->code == '1008')
                $card_number_warning .= (strlen($card_number_warning) > 0 ? ', ' : '') . $errorDetails->message;
            }

            if ($card_number_warning != '')
              $form_state->setErrorByName('payment_information][add_payment_method][payment_details][number', $card_number_warning);
            if ($cardholder_name_warning != '')
              $form_state->setErrorByName('payment_information][add_payment_method][payment_details][cardholder-name', $cardholder_name_warning);
            if ($card_cvv_warning != '')
              $form_state->setErrorByName('payment_information][add_payment_method][payment_details][security-code', $card_cvv_warning);
            if ($card_expiry_warning != '')
              $form_state->setErrorByName('payment_information][add_payment_method][payment_details][expiration', $card_expiry_warning);
            if ($token_problem)
              $form_state->setErrorByName('payment_information][add_payment_method][payment_details][number', $token_problem);

            if ($card_number_warning != '' || $cardholder_name_warning != '' || $card_cvv_warning != '' || $card_expiry_warning != '' || $token_problem != '') {
              $errorFeedback = true;
            }
          }
          else
          {
            $card_number_warning = '';
            $token_problem = '';
            foreach($card_identifier_errors as $errorDetails) {
              if ($errorDetails->code == '1000' || $errorDetails->code == '1001' || $errorDetails->code == '1002')
                $card_number_warning .= (strlen($card_number_warning) > 0 ? ', ' : '') . $errorDetails->message;
              else {
                $token_problem .= (strlen($token_problem) > 0 ? ', ' : '') . $errorDetails->code;
              }
            }

            if ($card_number_warning != '')
              $form_state->setErrorByName('payment_information][add_payment_method][payment_details][number', $card_number_warning);
            if ($token_problem)
              $form_state->setErrorByName('payment_information][add_payment_method][payment_details][number', $token_problem);

            if ($card_number_warning != '' || $token_problem != '') {
              $errorFeedback = true;
            }
          }
          if ($errorFeedback == false) {
            // Havent't got detailed feedback about issue, instead put out a generic message with the error codes
            $errorCodes = '';
            foreach ($card_identifier_errors as $errorDetails) {
              $errorCodes .= (strlen($errorCodes) > 0 ? ', ' : '') . $errorDetails->code;
            }
            $form_state->setErrorByName('payment_information][add_payment_method][payment_details][number', $this->t('Technical error validating your card details (HTTP code: ' . $card_identifier_http_error_code . ' , Error codes: ' . $errorCodes . ' at ' . $date_utc->format(\DateTime::RFC850) . ')'));
          }
        }
        else {
          // No card identifier but also no feedback from tokenisation
          $form_state->setErrorByName('payment_information][add_payment_method][payment_details][number', $this->t('Technical error validating your card details (' . $date_utc->format(\DateTime::RFC850) . ')'));
        }
      }

      // Do not call the parent's 'validateCreditCardForm' method (it would try to validate card details but these are not returned with the form)
    }

    return $element;
  }


  /**
   * Handles the submission of Opayo Pi payment details form.
   *
   * @param array $element
   *   The credit card form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   */
  protected function submitOpayoPiOwnForm(array $element, FormStateInterface &$form_state) {

    // Javascript should have filled in the 'card-identifier' and 'card-identifier-expiry' fields before submitting the form
    // Javascript manipulation of Drupal form fields is outside the scope of '$form_state', need to be copied explicitly
    $user_input = $form_state->getUserInput();
    try {
      $card_identifier = $user_input["payment_information"]["add_payment_method"]["payment_details"]["card-identifier"];
      $card_identifier_expiry = $user_input["payment_information"]["add_payment_method"]["payment_details"]["card-identifier-expiry"];
      $merchant_session_key = $user_input["payment_information"]["add_payment_method"]["payment_details"]["merchant-session-key"];
      $merchant_session_key_expiry = $user_input["payment_information"]["add_payment_method"]["payment_details"]["merchant-session-key-expiry"];
      $merchant_session_key_sequence_nr = $user_input["payment_information"]["add_payment_method"]["payment_details"]["merchant-session-key-sequence-nr"];
      $form_state->setValueForElement($element["card-identifier"], $card_identifier);
      $form_state->setValueForElement($element["card-identifier-expiry"], $card_identifier_expiry);
      $form_state->setValueForElement($element["merchant-session-key"], $merchant_session_key);
      $form_state->setValueForElement($element["merchant-session-key-expiry"], $merchant_session_key_expiry);
      $form_state->setValueForElement($element["merchant-session-key-sequence-nr"], $merchant_session_key_sequence_nr);
    } catch (Exception $e) {
    }

    // Browser settings collected by javascript to be used for 'strong customer authentication'
    try {
      $browserInfo = new stdClass();
      $browserInfo->colorDepth = 32;
      $browserInfo->screenHeight = 512;
      $browserInfo->screenWidth = 1024;
      $browserInfo->timeZoneOffset = 60;
      $browserInfo->language = 'en-GB';
      if (array_key_exists('browser-color-depth', $user_input))
        $browserInfo->colorDepth = $user_input['browser-color-depth'];
      if (array_key_exists('browser-screen-heighth', $user_input))
        $browserInfo->screenHeight = $user_input['browser-screen-height'];
      if (array_key_exists('browser-screen-width', $user_input))
        $browserInfo->screenWidth = $user_input['browser-screen-width'];
      if (array_key_exists('browser-timezone-offset', $user_input))
        $browserInfo->timeZoneOffset = $user_input['browser-timezone-offset'];
      if (array_key_exists('browser-language', $user_input))
        $browserInfo->language = $user_input['browser-language'];

      // Store it in the session so that it's accessible to the 'createPayment' method of the Payment Gateway
            /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin  */
      $pgplugin = $this->plugin;
      $order = $pgplugin->getOrder();
      $session = \Drupal::request()->getSession();
      $session->set('checkout-order-' . $order->id() . '-browser-extra-info', $browserInfo);

    } catch (Exception $e) {
    }
  }

}
