<?php

namespace Drupal\commerce_opayo_pi\PluginForm\OpayoPi;

use Drupal\commerce_payment\Exception\DeclineException;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\PluginForm\PaymentMethodEditForm as BasePaymentMethodEditForm;
use Drupal\commerce_payment\PluginForm\PaymentMethodFormBase;
use Drupal\Core\Form\FormStateInterface;
use DateTime;
use Drupal;
use Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGateway;
use Exception;
use stdClass;

class PaymentMethodEditForm extends BasePaymentMethodEditForm {

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {

    $form = PaymentMethodFormBase::buildConfigurationForm($form, $form_state);

    /** @var \Drupal\commerce_payment\Entity\PaymentMethodInterface $payment_method */
    $payment_method = $this->entity;

    $form['payment_details'] = [
      '#parents' => array_merge($form['#parents'], ['payment_details']),
      '#type' => 'container',
      '#payment_method_type' => $payment_method->bundle(),
    ];

    if ($payment_method->bundle() == 'credit_card_opayo') {
      /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin */
      $pgplugin = $this->plugin;
      $gateway_config = $pgplugin->getConfiguration();
      if ($gateway_config['form_type'] == 'dropin') {
        // Use drop-in version of the form
        $form['payment_details'] = $this->buildOpayoPiDropinForm($form['payment_details'], $form_state);
      }
      else {
        // Use the 'own form' version
        $form['payment_details'] = $this->buildOpayoPiOwnForm($form['payment_details'], $form_state);
    // Use an #after_build directive to tweak the payment details fields
        $form['payment_details']['#after_build'][] = [get_class($this), 'customisePaymentDetails'];
      }
    }

    // Use an #after_build directive to tweak the billing address fields
    if ($gateway_config['form_type'] == 'ownform') {
      // For 'own form', have cardholder first name and last name with the card details fields rather than with the address fields
      $form['billing_information']['address']['widget'][0]['address']['#after_build'][] = [get_class($this), 'customiseAddress'];
    }

    // Use an #after_build directive to tweak the billing information fields
    $form['billing_information']['#after_build'][] = [get_class($this), 'customiseBillingInfo'];


    // Move the billing information below the payment details.
    if (isset($form['billing_information'])) {
      $form['billing_information']['#weight'] = 10;
    }

    return $form;
  }

  /**
   * Tweak some aspects of the address fields for 'own form'
   */
  public static function customisePaymentDetails(array $element, FormStateInterface $form_state) {

    // We want the credit card field form elements to be displayed (and filled in by the user) but not get returned with the post (a tokenised 'card identifier' is returned instead)
    // Taking away the 'name' attribute should ensure this
    unset($element['cardholder-name']['#name']);
    unset($element['number']['#name']);
    unset($element['expiration']['month']['#name']);
    unset($element['expiration']['year']['#name']);
    unset($element['security-code']['#name']);
    // Opayo require the 'first name' and 'last name' when submitting the transaction so these fields are not excluded from the 'post'

    return $element;
  }

  /**
   * Tweak some aspects of the address fields
   */
  public static function customiseAddress(array $element, FormStateInterface $form_state) {

    // We already have the cardholder name field (both for 'drop-in' and 'own form') so don't need the first name, middle name and surname as well
    unset($element['container0']);
    unset($element['given_name']);
    unset($element['family_name']);
    unset($element['additional_name']);

    return $element;
  }

  /**
   * Tweak some aspects of the billing information
   */
  public static function customiseBillingInfo(array $element, FormStateInterface $form_state) {

    if ($form_state->getValue('process_input') == false) {
      // Match the phone number's country to the billing country

      // This is only effective on initial form build. When the user changes the country the Javascript only affects the 'address' section and
      // will not update the international telephone number field. Update in javascript instead.
      $countryCode = null;
      try {
        if (array_key_exists('address', $element) && array_key_exists('widget', $element['address']) && array_key_exists(0, $element['address']['widget']) && array_key_exists('address', $element['address']['widget'][0]) && array_key_exists('country_code', $element['address']['widget'][0]['address'])) {
          $countryCode = $element['address']['widget'][0]['address']['country_code']['#value'];
          if ($countryCode != null) {
            // Need to get the field name for the telephone number from the gateway configuration
            $payment_gateway_storage = \Drupal::entityTypeManager()->getStorage('commerce_payment_gateway');
            $payment_gateways = $payment_gateway_storage->loadByProperties(['status' => TRUE]);
            /** @var \Drupal\commerce_payment\Entity\PaymentGatewayInterface $payment_gateway */
            $payment_gateway = $payment_gateways['opayo_pi'];
            /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $plugin  */
            $plugin = $payment_gateway->getPlugin();
            $configuration = $plugin->getConfiguration();
            $telephoneFieldName = $configuration['telephone_field_name'];

            try {
              // If the telephone field is of type 'phone_international' (https://www.drupal.org/project/phone_international) then the following code copies the country code
              $element[$telephoneFieldName]['widget'][0]['value']['#country'] = $countryCode;
              $element[$telephoneFieldName]['widget'][0]['value']['int_phone']['#attributes']['data-country'] = $countryCode;
            }
            catch (Exception $ite) {}
          }
        }
      }
      catch (Exception $e) {
      }
    }

    return $element;
  }


  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    PaymentMethodFormBase::validateConfigurationForm($form, $form_state);
    /** @var \Drupal\commerce_payment\Entity\PaymentMethodInterface $payment_method */
    $payment_method = $this->entity;

    if ($payment_method->bundle() == 'credit_card_opayo') {
      /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin */
      $pgplugin = $this->plugin;
      $gateway_config = $pgplugin->getConfiguration();
      if ($gateway_config['form_type'] == 'dropin') {
        // Use drop-in version of the form
        $this->validateOpayoPiDropinForm($form['payment_details'], $form_state);
      }
      else {
        $this->validateOpayoPiOwnForm($form['payment_details'], $form_state);

        if ($form_state->isSubmitted())
        {
          // Need to copy the first name and last name from the 'payment details' section to the billing profile
          // (if it's done here the changes will get saved to the database - cannot wait until 'submitConfigurationForm')
          $cardHolderFirstname = $form_state->getValue($form["payment_details"]["cardholder"]["cardholder-first-name"]['#parents']);
          $cardHolderLastname = $form_state->getValue($form["payment_details"]["cardholder"]["cardholder-last-name"]['#parents']);
          $billingAddress = $form_state->getValue($form["billing_information"]["address"]["widget"][0]["address"]["#parents"]);
          $billingAddress['given_name'] = $cardHolderFirstname;
          $billingAddress['family_name'] = $cardHolderLastname;
          $form_state->setValueForElement($form["billing_information"]["address"]["widget"][0]["address"], $billingAddress);
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    PaymentMethodFormBase::submitConfigurationForm($form, $form_state);
    /** @var \Drupal\commerce_payment\Entity\PaymentMethodInterface $paymentMethod */
    $paymentMethod = $this->entity;
    /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin */
    $pgplugin = $this->plugin;

    if ($paymentMethod->bundle() == 'credit_card_opayo') {
      /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin */
      $pgplugin = $this->plugin;
      $gateway_config = $pgplugin->getConfiguration();
      if ($gateway_config['form_type'] == 'dropin') {
        // Use drop-in version of the form
        $this->submitOpayoPiDropinForm($form['payment_details'], $form_state);
      }
      else {
        // 'Own Form' version
        $this->submitOpayoPiOwnForm($form['payment_details'], $form_state);
      }
    }

    // Submitted checkout form values
    $formValues = $form_state->getValue($form['#parents']);

    // Add the order id
    $order = $pgplugin->getOrder();
    $formValues['payment_details']['order-id'] = $order->id();

    // The payment method form is customer facing. For security reasons
    // the returned errors need to be more generic.
    try {
      $pgplugin->createPaymentMethod($paymentMethod, $formValues['payment_details']);
    }
    catch (DeclineException $e) {
      $this->logger->warning($e->getMessage());
      throw new DeclineException(t('We encountered an error processing your payment method. Please verify your details and try again.'));
    }
    catch (PaymentGatewayException $e) {
      $this->logger->error($e->getMessage());
      throw new PaymentGatewayException(t('We encountered an unexpected error processing your payment method. Please try again later.'));
    }
  }


  /**
   * Builds the credit card form.
   *
   * @param array $element
   *   The target element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   *
   * @return array
   *   The built credit card form.
   */
  protected function buildOpayoPiDropinForm(array $element, FormStateInterface $form_state) {

    /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin  */
    $pgplugin = $this->plugin;
    $order = $pgplugin->getOrder();

    /** @var StdClass $merchantSessionKey  */
    $merchantSessionKey = $pgplugin->getValidMerchantSessionKey($order->id(), 0);

    // How many seconds until expiry
    $expiryinsecs = (new DateTime($merchantSessionKey->expiry))->getTimestamp() - (new DateTime('now'))->getTimestamp();

    $computed_settings = [
      'merchantSessionKey' => $merchantSessionKey->merchantSessionKey,
      'merchantSessionKeyExpirySecs' => $expiryinsecs,
      'merchantSessionKeyExpiry' => $merchantSessionKey->expiry,
      'merchantSessionKeySequenceNr' => $merchantSessionKey->sequenceNr,
      'orderId' => $order->id(),
    ];

    // Javascript that calls Opayo's 'sagepayCheckout' function through Drupal.behaviors
    $element['#attached']['library'][] = 'commerce_opayo_pi/opayo_pi_dropinform';
    // Pass the settings including the merchant session to the Javascript
    $element['#attached']['drupalSettings']['commerceopayopi']['opayoCheckout'] = $computed_settings;

    $gatewayConfig = $pgplugin->getConfiguration();
    $opayoJsRef = null;
    if($gatewayConfig['mode'] == 'test')
      $opayoJsRef = 'commerce_opayo_pi/opayo_sagepay_test';
    else
       $opayoJsRef = 'commerce_opayo_pi/opayo_sagepay_live';
    // Attach the external Opayo provided javascript
    $element['#attached']['library'][] = $opayoJsRef;

    // Feedback about expiry of the card fields
    $element['expiry-msg'] = [
      '#type' => 'markup',
      '#markup' => '<div id="expiry-msg" class="commerce-opayo-pi-expiry" style="font-size:smaller"></div>',
    ];

    // Hidden elements with the merchant session key info
    $element['merchant-session-key'] = [
      '#type' => 'hidden',
      '#value' => $merchantSessionKey->merchantSessionKey,
      '#attributes' => [
        'id' => ['merchant-session-key'],
      ],
    ];
    $element['merchant-session-key-sequence-nr'] = [
      '#type' => 'hidden',
      '#value' => $merchantSessionKey->sequenceNr,
      '#attributes' => [
        'id' => ['merchant-session-key-sequence-nr'],
      ],
    ];
    $element['merchant-session-key-expiry'] = [
      '#type' => 'hidden',
      '#value' => $merchantSessionKey->expiry,
      '#attributes' => [
        'id' => ['merchant-session-key-expiry'],
      ],
    ];

    // HTML element that will contain the iFrame for Opayo's 'Drop-In' elements
    $element['sp-container'] = [
      '#type' => 'container',
      '#attributes' => [
        'id' => ['sp-container'],
      ],
    ];

    return $element;
  }


  /**
   * Validates the credit card form.
   *
   * @param array $element
   *   The credit card form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   */
  protected function validateOpayoPiDropinForm(array &$element, FormStateInterface $form_state) {

    if ($form_state->isSubmitted())
    {
      $cardIdentifier = '';
      $cardIdentifierHttpCode = 0;
      $cardIdentifierErrorCode = 0;
      $cardIdentifierErrorMessage = '';

      if (array_key_exists('card-identifier', $form_state->getUserInput()))
        // Opayo's 'Drop-In' logic should have provided a 'Card Identifier' among the posted fields
        $cardIdentifier = $form_state->getUserInput()['card-identifier'];

      if (!$cardIdentifier) {
        // Check to see if we had an expired merchant session key

        /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGateway $pgplugin  */
        $pgplugin = $this->plugin;
        $currentMerchantSessionKey = $pgplugin->getCurrentMerchantSessionKey($pgplugin->getOrder()->id());
        if ($currentMerchantSessionKey != null)
        {
          if (new DateTime('now') >= new DateTime($currentMerchantSessionKey->expiry))
            $form_state->setErrorByName('sp-container', $this->t('Your card details could not be validated.'));
        }

        // Check for any error fields in the POST that were supplied by the Opayo javascript
        if (array_key_exists('card-identifier-http-code', $form_state->getUserInput()))
          $cardIdentifierHttpCode = $form_state->getUserInput()['card-identifier-http-code'];
        if (array_key_exists('card-identifier-error-code', $form_state->getUserInput()))
          $cardIdentifierErrorCode = $form_state->getUserInput()['card-identifier-error-code'];
        if (array_key_exists('card-identifier-error-message', $form_state->getUserInput()))
          $cardIdentifierErrorMessage = $form_state->getUserInput()['card-identifier-error-message'];

        if (isset($cardIdentifierHttpCode) && isset($cardIdentifierErrorCode) && isset($cardIdentifierErrorMessage))
        {
          // TODO: refine
          $form_state->setErrorByName('sp-container', 'HTTP Code: ' . $cardIdentifierHttpCode . ', Error code: ' . $cardIdentifierErrorCode . ', Error message: ' . $cardIdentifierErrorMessage);
        }
        else {
          // Display a generic error
          $form_state->setErrorByName('sp-container', $this->t('Your card details could not be validated.'));
        }

        // Invalidate the current merchant session key
        /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OnsiteOpayoPiDropinGateway $pgplugin  */
        $pgplugin = $this->plugin;
        $pgplugin->invalidateCurrentMerchantSessionKey();
        return;
      }

      // Do not call the parent method (it would try to validate card details that are handled by the drop-in form instead)
    }
  }

  /**
   * Handles the submission of the credit card form.
   *
   * @param array $element
   *   The credit card form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   */
  protected function submitOpayoPiDropinForm(array $element, FormStateInterface &$form_state) {
    // Opayo's 'Drop-In' logic should have provided a 'Card Identifier' among the posted fields
    $cardIdentifier = $form_state->getUserInput()['card-identifier'];

    // Update the form_state to contain the payment details, in this case the only payment detail is the 'card_identifier' token
    $paymentDetails = [ 'card-identifier' => $cardIdentifier ];
    $form_state->setValueForElement($element, $paymentDetails);

    // Current merchant session key has been used succesfully to with the card details to generate a 'card identifier'. It cannot be
    // used again, invalidate it
    /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OnsiteOpayoPiDropinGateway $pgplugin  */
    $pgplugin = $this->plugin;
    $pgplugin->invalidateCurrentMerchantSessionKey();

  }

  /**
   * Builds the credit card form.
   *
   * @param array $element
   *   The target element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   *
   * @return array
   *   The built credit card form.
   */
  protected function buildOpayoPiOwnForm(array $element, FormStateInterface $form_state) {

    // Credit card fields
    $element = $this->buildOpayoCreditCardForm($element, $form_state);

    /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin  */
    $pgplugin = $this->plugin;
    $order = $pgplugin->getOrder();

    /** @var StdClass $merchantSessionKey  */
    $merchantSessionKey = $pgplugin->getValidMerchantSessionKey($order->id(), 120);

    // How many seconds until expiry
    $expiryinsecs = (new DateTime($merchantSessionKey->expiry))->getTimestamp() - (new DateTime('now'))->getTimestamp();

    $computed_settings = [
      'merchantSessionKey' => $merchantSessionKey->merchantSessionKey,
      'merchantSessionKeyExpirySecs' => $expiryinsecs,
      'merchantSessionKeyExpiry' => $merchantSessionKey->expiry,
      'merchantSessionKeySequenceNr' => $merchantSessionKey->sequenceNr,
      'orderId' => $order->id(),
    ];

    // Javascript that calls 'sagepayOwnForm' through Drupal.behaviors
    $element['#attached']['library'][] = 'commerce_opayo_pi/opayo_pi_ownform';
    // Pass the settings including the merchant session to the Javascript
    $element['#attached']['drupalSettings']['commerceopayopi']['opayoCheckout'] = $computed_settings;

    $gatewayConfig = $pgplugin->getConfiguration();
    $opayoJsRef = null;
    if ($gatewayConfig['mode'] == 'test')
    $opayoJsRef = 'commerce_opayo_pi/opayo_sagepay_test';
    else
      $opayoJsRef = 'commerce_opayo_pi/opayo_sagepay_live';
    // Attach the external Opayo provided javascript
    $element['#attached']['library'][] = $opayoJsRef;

    // hidden element for the card identifier pass-back
    $element['card-identifier'] = [
      '#type' => 'hidden',
      '#value' => '',
      '#attributes' => [
        'id' => ['card-identifier'],
      ],
    ];
    $element['card-identifier-expiry'] = [
      '#type' => 'hidden',
      '#value' => '',
      '#attributes' => [
        'id' => ['card-identifier-expiry'],
      ],
    ];
    $element['card-identifier-result'] = [
      '#type' => 'hidden',
      '#value' => '',
      '#attributes' => [
        'id' => ['card-identifier-result'],
      ],
      '#required' => TRUE,
    ];
    $element['card-identifier-error-httpcode'] = [
      '#type' => 'hidden',
      '#value' => '',
      '#attributes' => [
        'id' => ['card-identifier-error-httpcode'],
      ],
    ];
    // Hidden elements with the merchant session key info
    $element['merchant-session-key'] = [
      '#type' => 'hidden',
      '#value' => $merchantSessionKey->merchantSessionKey,
      '#attributes' => [
        'id' => ['merchant-session-key'],
      ],
    ];
    $element['merchant-session-key-sequence-nr'] = [
      '#type' => 'hidden',
      '#value' => $merchantSessionKey->sequenceNr,
      '#attributes' => [
        'id' => ['merchant-session-key-sequence-nr'],
      ],
    ];
    $element['merchant-session-key-expiry'] = [
      '#type' => 'hidden',
      '#value' => $merchantSessionKey->expiry,
      '#attributes' => [
        'id' => ['merchant-session-key-expiry'],
      ],
    ];

    return $element;
  }

  /**
   * Builds the credit card form.
   *
   * @param array $element
   *   The target element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   *
   * @return array
   *   The built credit card form.
   */

  protected function buildOpayoCreditCardForm(array $element, FormStateInterface $form_state) {

    /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin */
      $pgplugin = $this->plugin;
      $gateway_config = $pgplugin->getConfiguration();

    // Build a month select list that shows months with a leading zero.
    $months = [];
    for ($i = 1; $i < 13; $i++) {
      $month = str_pad($i, 2, '0', STR_PAD_LEFT);
      $months[$month] = $month;
    }
    // Build a year select list that uses a 4 digit key with a 2 digit value.
    $current_year_4 = date('Y');
    $current_year_2 = date('y');
    $years = [];
    for ($i = 0; $i < 10; $i++) {
      $years[$current_year_4 + $i] = $current_year_2 + $i;
    }

    $element['#attributes']['class'][] = 'credit-card-form';
    // Placeholder for the detected card type. Set by validateCreditCardForm().
    $element['type'] = [
      '#type' => 'hidden',
      '#value' => '',
    ];
    // Put cardholder first and last name with the card details
    $element['cardholder'] = [
      '#type' => 'container',
      '#attributes' => [
        'class' => ['credit-card-form__cardholder'],
      ],
    ];
    $element['cardholder']['cardholder-first-name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('First Name'),
      '#attributes' => ['autocomplete' => 'off', 'data-card-details' => 'cardholder-first-name'],
      '#required' => TRUE,
      '#maxlength' => 35,
      '#size' => 12,
    ];
    // Allow the surname to stay blank in 'test' mode so that the first name can be used for Opayo 'Magic CardHolder Values'
    $element['cardholder']['cardholder-last-name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Last Name'),
      '#attributes' => ['autocomplete' => 'off', 'data-card-details' => 'cardholder-last-name'],
      '#required' => TRUE,
      '#maxlength' => 35,
      '#size' => 12,
    ];
    $element['cardholder-name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Card Holder'),
      '#attributes' => ['autocomplete' => 'off', 'data-card-details' => 'cardholder-name'],
      '#required' => TRUE,
      '#maxlength' => 35,
      '#size' => 25,
    ];
    $element['number'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Card number'),
      '#attributes' => ['autocomplete' =>'off', 'data-card-details' => 'card-number'],
      '#required' => TRUE,
      '#maxlength' => 19,
      '#size' => 25,
    ];
    $element['expiration'] = [
      '#type' => 'container',
      '#attributes' => [
        'class' => ['credit-card-form__expiration'],
      ],
    ];
    $element['expiration']['month'] = [
      '#type' => 'select',
      '#title' => $this->t('Expiry Month'),
      '#attributes' => ['data-card-details' => 'expiry-date-month'],
      '#options' => $months,
      '#default_value' => date('m'),
      '#required' => TRUE,
    ];
    $element['expiration']['year'] = [
      '#type' => 'select',
      '#title' => $this->t('Expiry Year'),
      '#attributes' => ['data-card-details' => 'expiry-date-year'],
      '#options' => $years,
      '#default_value' => $current_year_4,
      '#required' => TRUE,
    ];
    $element['security-code'] = [
      '#type' => 'textfield',
      '#title' => $this->t('CVV'),
      '#attributes' => ['autocomplete' => 'off', 'data-card-details' => 'security-code'],
      '#required' => TRUE,
      '#maxlength' => 4,
      '#size' => 4,
    ];

    return $element;
  }

  /**
   * Validates the credit card form when using 'Own Form' appreach.
   *
   * @param array $element
   *   The credit card form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   */
  protected function validateOpayoPiOwnForm(array &$element, FormStateInterface $form_state) {

    if ($form_state->isSubmitted())
    {
      // Checks needed when payment details have been submitted

      /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin  */
      $pgplugin = $this->plugin;

      $cardIdentifier = null;
      $merchantSessionKey = null;
      $cardIdentifierHttpErrorCode = null;
      $cardIdentifierErrors = array();

      // The values, including 'payment_details', that are added to the form by the javascript are not found in the $form_state's 'values' but
      // can be retrieved from its 'input' field.

      $userInput = $form_state->getUserInput();
      try {
        $result = $userInput["payment_information"]["add_payment_method"]["payment_details"]["card-identifier-result"];
        $merchantSessionKey = $userInput["payment_information"]["add_payment_method"]["payment_details"]["merchant-session-key"];
        if ($result == "true")
        {
          $cardIdentifier = $userInput["payment_information"]["add_payment_method"]["payment_details"]["card-identifier"];
        }
        else {
          $cardIdentifierHttpErrorCode = $userInput["payment_information"]["add_payment_method"]["payment_details"]["card-identifier-error-httpcode"];

          $index = 0;
          if (array_key_exists("card-identifier-error-" . $index . "-code", $userInput))
          {
            $errorDetails = new stdClass();
            $errorDetails->code = $userInput["card-identifier-error-" . $index . "-code"];
            $errorDetails->message = $userInput["card-identifier-error-" . $index . "-message"];
            array_push($cardIdentifierErrors, $errorDetails);
            $index++;
          }
        }
      }
      catch (Exception $e) {
      }

      if ($form_state->hasAnyErrors()) {
        // Some error messages are the result of the fact that the credit card fields are on the form but their values are tokenised using javascript
        // and not returned here. Filter those out.
        $formErrors = $form_state->getErrors();
        $filteredFormErrors = array();
        /** @var Drupal\Core\StringTranslation\TranslatableMarkup $errorElement */
        foreach ($formErrors as $errorElement => $errorObject) {
          $ignore = (strval($errorElement) == 'payment_information][add_payment_method][payment_details][cardholder][cardholder-first-name' ||
            strval($errorElement) == 'payment_information][add_payment_method][payment_details][cardholder][cardholder-last-name' ||
            strval($errorElement) == 'payment_information][add_payment_method][payment_details][cardholder-name' ||
            strval($errorElement) == 'payment_information][add_payment_method][payment_details][number' ||
            strval($errorElement) == 'payment_information][add_payment_method][payment_details][security-code');
          if (!$ignore) {
            $filteredFormErrors[$errorElement] = $errorObject;
          }
        }
        if (sizeof($filteredFormErrors) < sizeof($formErrors)) {
          // Some errors have been filtered out - reset
          $form_state->clearErrors();
          foreach ($filteredFormErrors as $errorElement => $errorObject) {
            $form_state->setErrorByName($errorElement, $errorObject);
          }
        }
      }

      if ($cardIdentifier != null) {
        // Successful tokenisation of card details has taken place

        if ($form_state->hasAnyErrors()) {
          $order = $pgplugin->getOrder();
          // There are still errors - not relating to the card detail fields
          // The merchant session key has been used and cannot be reused.
          // Use the hidden fields to pass back a fresh merchant session key
          /** @var StdClass $newMerchantSessionKey  */
          $newMerchantSessionKey = $pgplugin->getValidMerchantSessionKey($order->id(), 0);
          $form_state->setValue($element["merchant-session-key"]["#parents"], $newMerchantSessionKey->merchantSessionKey);
          $form_state->setValue($element["merchant-session-key-sequence-nr"]["#parents"], $newMerchantSessionKey->sequenceNr);
          $form_state->setValue($element["merchant-session-key-expiry"], $newMerchantSessionKey->expiry);
        }
      }
      else {
        // No card-identifier returned with the form

        $date_utc = new \DateTime("now", new \DateTimeZone("UTC"));
        if ($cardIdentifierHttpErrorCode != null) {
          $errorFeedback = false;
          if ($cardIdentifierHttpErrorCode == 422)
          {
            $cardNumberWarning = '';
            $cardExpiryWarning = '';
            $cardCVVWarning = '';
            $tokenProblem = '';
            foreach($cardIdentifierErrors as $errorDetails) {
              if (str_contains(strtolower($errorDetails->message), 'card number')) {
                $cardNumberWarning .= (strlen($cardNumberWarning) > 0 ? ', ' : '') . $errorDetails->message;
              }
              if ($errorDetails->code == '1006' || $errorDetails->code == '1011')
                $tokenProblem .= (strlen($tokenProblem) > 0 ? ', ' : '') . $errorDetails->code;
              if ($errorDetails->code == '1008')
                $cardNumberWarning .= (strlen($cardNumberWarning) > 0 ? ', ' : '') . $errorDetails->message;
            }

            if ($cardNumberWarning != '')
              $form_state->setErrorByName('payment_information][add_payment_method][payment_details][number', $cardNumberWarning);
            if ($cardCVVWarning != '')
              $form_state->setErrorByName('payment_information][add_payment_method][payment_details][security-code', $cardCVVWarning);
            if ($cardExpiryWarning != '')
              $form_state->setErrorByName('payment_information][add_payment_method][payment_details][expiration', $cardExpiryWarning);
            if ($tokenProblem)
              $form_state->setErrorByName('payment_information][add_payment_method][payment_details][number', $tokenProblem);

            if ($cardNumberWarning != '' || $cardCVVWarning != '' || $cardExpiryWarning != '' || $tokenProblem = '') {
              $errorFeedback = true;
            }
          }
          if ($errorFeedback == false) {
            // Havent't got detailed feedback about issue, instead put out a generic message with the error codes
            $errorCodes = '';
            foreach ($cardIdentifierErrors as $errorDetails) {
              $errorCodes .= (strlen($errorCodes) > 0 ? ', ' : '') . $errorDetails->code;
            }
            $form_state->setErrorByName('payment_information][add_payment_method][payment_details][number', $this->t('Technical error validating your card details (HTTP code: ' . $cardIdentifierHttpErrorCode . ' , Error codes: ' . $errorCodes . ' at ' . $date_utc->format(\DateTime::RFC850) . ')'));
          }

          // The merchant session key has been used and cannot be reused.
          // Use the hidden fields to pass back a fresh merchant session key
          /** @var StdClass $newMerchantSessionKey  */
          $order = $pgplugin->getOrder();
          $newMerchantSessionKey = $pgplugin->getValidMerchantSessionKey($order->id(), 0);
          // When the form 'returns' it will not be rebuilt based on the '$form_state', irrespective of whether the 'rebuild' flag is set
          // (logic in 'processForm' in the Core 'FormBuilder')
          // Need to update the values in the $element instead
          $element["merchant-session-key"]["#value"] = $newMerchantSessionKey->merchantSessionKey;
          $element["merchant-session-key-sequence-nr"]["#value"] = $newMerchantSessionKey->sequenceNr;
          $element["merchant-session-key-expiry"]["#value"] = $newMerchantSessionKey->expiry;
        }
        else {
          // No card identifier but also no feedback from tokenisation
          $form_state->setErrorByName('payment_information][add_payment_method][payment_details][number', $this->t('Technical error validating your card details (' . $date_utc->format(\DateTime::RFC850) . ')'));
        }
      }

      // The merchant session key has been used, cannot be reused
      $pgplugin->invalidateMerchantSessionKey($pgplugin->getOrder()->id(), $merchantSessionKey);

      // Do not call the parent's 'validateCreditCardForm' method (it would try to validate card details but these are not returned with the form)
    }

    return $element;
  }


  /**
   * Handles the submission of Opayo Pi payment details form.
   *
   * @param array $element
   *   The credit card form element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the complete form.
   */
  protected function submitOpayoPiOwnForm(array $element, FormStateInterface &$form_state) {

    // Javascript should have filled in the 'card-identifier' and 'card-identifier-expiry' fields before submitting the form
    // Javascript manipulation of Drupal form fields is outside the scope of '$form_state', need to be copied explicitly
    $userInput = $form_state->getUserInput();
    try {
      $cardIdentifier = $userInput["payment_information"]["add_payment_method"]["payment_details"]["card-identifier"];
      $cardIdentifierExpiry = $userInput["payment_information"]["add_payment_method"]["payment_details"]["card-identifier-expiry"];
      $form_state->setValueForElement($element["card-identifier"], $cardIdentifier);
      $form_state->setValueForElement($element["card-identifier-expiry"], $cardIdentifierExpiry);
    } catch (Exception $e) {
    }

    // Browser settings collected by javascript to be used for 'strong customer authentication'
    try {
      $browserInfo = new stdClass();
      $browserInfo->colorDepth = 32;
      $browserInfo->screenHeight = 512;
      $browserInfo->screenWidth = 1024;
      $browserInfo->timeZoneOffset = 60;
      $browserInfo->language = 'en-GB';
      if (array_key_exists('browser-color-depth', $userInput))
        $browserInfo->colorDepth = $userInput['browser-color-depth'];
      if (array_key_exists('browser-screen-heighth', $userInput))
        $browserInfo->screenHeight = $userInput['browser-screen-height'];
      if (array_key_exists('browser-screen-width', $userInput))
        $browserInfo->screenWidth = $userInput['browser-screen-width'];
      if (array_key_exists('browser-timezone-offset', $userInput))
        $browserInfo->timeZoneOffset = $userInput['browser-timezone-offset'];
      if (array_key_exists('browser-language', $userInput))
        $browserInfo->language = $userInput['browser-language'];

      // Store it in the session so that it's accessible to the 'createPayment' method of the Payment Gateway
            /** @var \Drupal\commerce_opayo_pi\Plugin\Commerce\PaymentGateway\OpayoPiPaymentGatewayInterface $pgplugin  */
      $pgplugin = $this->plugin;
      $order = $pgplugin->getOrder();
      $session = \Drupal::request()->getSession();
      $session->set('checkout-order-' . $order->id() . '-browser-extra-info', $browserInfo);

    } catch (Exception $e) {
    }
  }

}


